using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Xml;
using Hardcodet.Scheduling;
using NUnit.Framework;


namespace Scheduling_Tests
{
  /// <summary>
  /// 
  /// </summary>
  [TestFixture]
  public class Given_Job_When_Pausing_Execution : SchedulerTestBase
  {
    private Job job;

    protected override void InitInternal()
    {
      job = new Job("abc");
      Scheduler.MinJobInterval = 1;
    }


    [Test]
    public void Pausing_Should_Prevent_Execution()
    {
      int counter = 0;
      
      job.Run.Every.Milliseconds(100).Until(DateTime.Now.AddMilliseconds(500));
      Scheduler.SubmitJob(job, j => counter++);

      Assert.IsTrue(job.Pause());
      Assert.AreEqual(JobState.Paused, job.State);

      Thread.Sleep(200);
      Assert.AreEqual(0, counter);
    }

    [Test]
    public void Pausing_Via_Scheduler_Should_Work()
    {
      int counter = 0;

      job.Run.Every.Milliseconds(100).Until(DateTime.Now.AddMilliseconds(500));
      Scheduler.SubmitJob(job, j => counter++);

      Assert.IsTrue(Scheduler.PauseJob(job.JobId));
      Assert.AreEqual(JobState.Paused, job.State);

      Thread.Sleep(200);
      Assert.AreEqual(0, counter);
    }




    [ExpectedException(typeof(InvalidOperationException))]
    [Test]
    public void Pausing_Job_Without_Interval_Should_Fail()
    {
      //if there is no interval, we cannot reschedule
      job.Run.From(DateTime.Now.AddMinutes(1)).Once();
      Scheduler.SubmitJob(job, j => { });

      job.Pause();
    }

    [Test]
    public void Paused_Jobs_Should_Not_Decrease_Loop_Counter()
    {
      int counter = 0;

      //if there is no interval, we cannot reschedule
      job.Run.Times(10).Every.Milliseconds(100);
      Scheduler.SubmitJob(job, j => counter++);

      Thread.Sleep(250);
      Assert.AreEqual(3, counter);

      job.Pause();
      Thread.Sleep(1000);
      Assert.AreEqual(3, counter);

      job.Resume();
      Thread.Sleep(1000);
      Assert.AreEqual(10, counter);
    }

    [Test]
    public void Paused_Jobs_Should_Be_Rescheduled_With_Each_Execution_Time()
    {
      int counter = 0;
      job.Run.Every.Seconds(2).Times(10);

      //if there is no interval, we cannot reschedule
      job.Run.Times(10).Every.Milliseconds(100);
      Scheduler.SubmitJob(job, j => counter++);

      Thread.Sleep(250);
      Assert.AreEqual(3, counter);

      job.Pause();

      //go forward - as if the jobs had been paused for ages
      //(remark: going back would need system time change detection and rescheduling)
      var now = DateTime.Now.AddHours(2.5);
      SystemTime.Now = () => now;

      Thread.Sleep(200);
      job.Resume();
      Thread.Sleep(1000);
      Assert.AreEqual(10, counter);
    }


    [Test]
    public void Pausing_Should_Not_Work_For_Canceled_Jobs ()
    {
      job.Run.Times(1).Every.Milliseconds(100);
      job.Cancel();
      Assert.IsFalse(job.Pause());
      Assert.AreEqual(JobState.Canceled, job.State);
    }


    [Test]
    public void Pausing_Should_Not_Work_For_Finished_Jobs()
    {
      //if there is no interval, we cannot reschedule
      job.Run.Times(1).Every.Milliseconds(100);
      Scheduler.SubmitJob(job, j => { });

      Thread.Sleep(110);
      Assert.AreEqual(JobState.Finished, job.State);
      Assert.IsFalse(job.Pause());
      Assert.AreEqual(JobState.Finished, job.State);
    }


    [Test]
    public void Paused_Jobs_Should_Still_Expire()
    {
      int counter = 0;
      //if there is no interval, we cannot reschedule
      job.Run.Every.Milliseconds(100).Until(DateTime.Now.AddMilliseconds(500));
      Scheduler.SubmitJob(job, j => counter++);

      Thread.Sleep(50);
      Assert.AreEqual(1, counter);
      Assert.AreEqual(JobState.Active, job.State);

      //now pause
      job.Pause();
      Assert.AreEqual(JobState.Paused, job.State);

      Thread.Sleep(500);

      //no more invocations, and the job should have finished
      Assert.AreEqual(1, counter);
      Assert.AreEqual(JobState.Finished, job.State);
    }
  }
}