using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Xml;
using Hardcodet.Scheduling;
using NUnit.Framework;


namespace Scheduling_Tests
{
  /// <summary>
  /// 
  /// </summary>
  [TestFixture]
  public class Given_Scheduler_When_Rescheduling_Due_To_Changed_System_Time : SchedulerTestBase
  {
    private DateTimeOffset initial, later, earlier;

    protected override void InitInternal()
    {
      initial = DateTimeOffset.Now;
      later = initial.AddMinutes(90);
      earlier = initial.AddMinutes(-120);

      SystemTime.Now = () => initial;
    }


    [Test]
    public void Pending_Jobs_Should_Run_Earlier_If_System_Time_Is_Moved_Back_To_Past()
    {
      int callbackCounter = 0;
      Scheduler.SystemTimeChangeRescheduling = ReschedulingStrategy.RescheduleNextExecution;
      Job job = new Job();

      //run 3 times with a 1 sec interval. start in half a sec
      job.Run.From(initial.AddSeconds(0.5)).Every.Seconds(1).Times(3);
      Scheduler.SubmitJob(job, j => callbackCounter++);

      //it's suddenly much earlier
      SystemTime.Now = () => earlier;

      //give scheduler time to run once
      Thread.Sleep(600);

      //the first job should have run anyway
      Assert.AreEqual(1, callbackCounter);

      //next execution time should be in half a sec
      Assert.AreEqual(earlier.AddSeconds(1), Scheduler.NextExecution);

     //the job is of course still active
      Assert.AreEqual(JobState.Active, job.State);
    }


    [Test]
    public void Pending_Jobs_Should_Not_All_Become_Due_If_System_Time_Is_Moved_Back_To_Past_With_Rescheduling_Enabled()
    {
      int callbackCounter = 0;
      Scheduler.SystemTimeChangeRescheduling = ReschedulingStrategy.RescheduleNextExecution;
      Job job = new Job();

      //run 3 times with a 1 sec interval. start in half a sec
      job.Run.From(initial.AddSeconds(0.5)).Every.Seconds(1).Times(3);
      Scheduler.SubmitJob(job, j => callbackCounter++);

      //it's suddenly much later
      SystemTime.Now = () => later;

      //give scheduler time to run
      Thread.Sleep(600);

      //the first job should have run anyway
      Assert.AreEqual(1, callbackCounter);

      //next execution time should be in half a sec
      Assert.AreEqual(later.AddSeconds(1), Scheduler.NextExecution);

      //the job is of course still active
      Assert.AreEqual(JobState.Active, job.State);
    }


    [Test]
    public void Jobs_That_Become_Far_In_The_Future_Due_To_Time_Shift_Should_Run_Later_If_Working_With_Absolute_Times()
    {
      int callbackCounter = 0;
      Scheduler.SystemTimeChangeRescheduling = ReschedulingStrategy.KeepFixedTimes;
      Scheduler.SelfTestInterval = int.MaxValue;

      Job job = new Job();

      //run 3 times with a 1 sec interval. start in half a sec
      job.Run.From(initial.AddSeconds(0.5)).Every.Seconds(1).Times(3).Until(initial.AddMinutes(5));
      Scheduler.SubmitJob(job, j => callbackCounter++);

      //it's suddenly much earlier
      SystemTime.Now = () => earlier;

      //give scheduler time to run
      Thread.Sleep(600);

      //the first job should not have run - it's way in the future
      Assert.AreEqual(0, callbackCounter);

      //next execution time should still be the job's original start time
      Assert.AreEqual(job.StartTime, Scheduler.NextExecution);

      //the job is of course still active
      Assert.AreEqual(JobState.Active, job.State);
    }



    [Test]
    public void Jobs_That_Expire_Due_To_Time_Shift_Should_Not_Run_If_Working_With_Absolute_Times()
    {
      int callbackCounter = 0;
      Scheduler.SystemTimeChangeRescheduling = ReschedulingStrategy.KeepFixedTimes;
      Scheduler.SelfTestInterval = int.MaxValue;

      Job job = new Job();

      //run 3 times with a 1 sec interval. start in half a sec
      job.Run.From(initial.AddSeconds(0.5)).Every.Seconds(1).Times(3).Until(initial.AddMinutes(5));
      Scheduler.SubmitJob(job, j => callbackCounter++);

      //it's suddenly much later
      SystemTime.Now = () => later;

      //give scheduler time to run
      Thread.Sleep(600);

      //the first job should not have run - it's way in the future
      Assert.AreEqual(0, callbackCounter);

      //next execution time should still be the job's original start time
      Assert.AreNotEqual(job.StartTime, Scheduler.NextExecution);

      //the job expired
      Assert.AreEqual(JobState.Finished, job.State);
    }


    [Test]
    public void Jobs_That_Expire_Due_To_Time_Shift_Should_Run_If_Expiration_Time_Is_Shifted()
    {
      int callbackCounter = 0;
      Scheduler.SystemTimeChangeRescheduling = ReschedulingStrategy.RescheduleNextExecutionAndExpirationTime;
      Job job = new Job();

      //run 3 times with a 1 sec interval. start in half a sec
      job.Run.From(initial.AddSeconds(0.5)).Every.Seconds(1).Times(3).Until(initial.AddMinutes(5));
      Scheduler.SubmitJob(job, j => callbackCounter++);

      //the expiration time is below the later time
      Assert.IsTrue(job.ExpirationTime < later);

      //it's suddenly much later
      SystemTime.Now = () => later;

      //give scheduler time to run
      Thread.Sleep(600);

      //the first job should have run anyway
      Assert.AreEqual(1, callbackCounter);

      //the expiration time should have shifted
      Assert.IsTrue(job.ExpirationTime > later);

      //next execution time should be in half a sec
      Assert.AreEqual(later.AddSeconds(1), Scheduler.NextExecution);

      //the job is of course still active
      Assert.AreEqual(JobState.Active, job.State);
    }

  }
}
