using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Xml;
using Hardcodet.Scheduling;
using NUnit.Framework;


namespace Scheduling_Tests
{
  /// <summary>
  /// 
  /// </summary>
  [TestFixture]
  public class Given_Scheduler_When_Submitting_Job : SchedulerTestBase
  {

    [Test]
    public void Submitting_First_Job_Should_Update_Execution_Time_Immediately()
    {
      //configure system time to return a fixed time
      DateTimeOffset now = DateTime.Now;
      SystemTime.Now = () => now;

      //schedule a new job that runs in 10 seconds
      DateTimeOffset start = now.AddSeconds(10);
      Job job = new Job();
      job.Run.From(start).Once();

      //submit the job
      Scheduler.SubmitJob(job, j => { });

      Assert.AreEqual(start, Scheduler.NextExecution);
    }


    [Test]
    public void Submitting_Job_With_Earlier_Due_Time_Should_Cause_Rescheduling()
    {
      //configure system time to return a fixed time
      DateTimeOffset now = DateTime.Now;
      SystemTime.Now = () => now;

      //schedule a new job that runs in 10 seconds
      DateTimeOffset start = now.AddSeconds(10);
      Job job1 = new Job();
      job1.Run.From(start).Once();
      Scheduler.SubmitJob(job1, j => { });
      Assert.AreEqual(start, Scheduler.NextExecution);

      Job job2 = new Job();
      job2.Run.From(now.AddSeconds(4)).Once();
      Scheduler.SubmitJob(job2, j => { });
      Assert.AreEqual(now.AddSeconds(4), Scheduler.NextExecution);
    }


    [Test]
    public void Submitting_Job_With_Later_Due_Time_Should_Keep_Schedule()
    {
      //configure system time to return a fixed time
      DateTimeOffset now = DateTime.Now;
      SystemTime.Now = () => now;

      //schedule a new job that runs in 10 seconds
      DateTimeOffset start = now.AddSeconds(10);
      Job job1 = new Job();
      job1.Run.From(start).Once();
      Scheduler.SubmitJob(job1, j => { });
      Assert.AreEqual(start, Scheduler.NextExecution);

      Job job2 = new Job();
      job2.Run.From(now.AddSeconds(44)).Once();
      Scheduler.SubmitJob(job2, j => { });


      Assert.AreEqual(start, Scheduler.NextExecution);
    }


    [Test]
    public void Submitting_Second_Job_With_Equal_Due_Time_Should_Keep_Schedule_And_Execute_Both_Jobs()
    {
      //configure system time to return a fixed time
      DateTimeOffset now = DateTimeOffset.Now;

      int counter = 0;

      //schedule a new job that runs in 2 seconds
      DateTimeOffset start = now.AddSeconds(2);
      Job job1 = new Job();
      job1.Run.From(start).Once();
      Scheduler.SubmitJob(job1, j => counter++);

      Job job2 = new Job();
      job2.Run.From(start).Once();
      Scheduler.SubmitJob(job2, j => counter++);

      Thread.Sleep(2050);
      Assert.AreEqual(2, counter);

      Assert.AreEqual(JobState.Finished, job1.State);
      Assert.AreEqual(JobState.Finished, job2.State);
    }

    [Test]
    public void Submitted_Jobs_Can_Be_Retrieved_Via_Their_Id()
    {
      var list = new List<Job>();
      for (int i = 0; i < 10; i++)
      {
        DateTimeOffset start = DateTime.Now.AddSeconds(10);
        Job job = new Job();
        job.Run.From(start).Once();

        //submit the job
        //submit the job
        Scheduler.SubmitJob(job, j => { });
        list.Add(job);
      }


      Assert.AreEqual(10, list.Count);
      foreach (var job in list)
      {
        Assert.AreSame(job, Scheduler.TryGetJob(job.JobId));
      }
      
    }
  }
}
